@@ -23,11 +23,10 @@
 
 use std::{
     convert::{TryFrom, TryInto},
-    fmt::{self, Debug, Display},
-    ops::{Deref, DerefMut},
+    fmt::{self, Debug, Display, Formatter},
 };
 
-use chrono::{Datelike, SecondsFormat, TimeZone, Utc};
+use chrono::Datelike;
 use serde_json::{json, Value};
 
 pub use crate::document::Document;
@@ -38,7 +37,7 @@ use crate::{
 };
 
 /// Possible BSON value types.
-#[derive(Clone, Debug, PartialEq)]
+#[derive(Clone, PartialEq)]
 pub enum Bson {
     /// 64-bit binary floating point
     Double(f64),
@@ -69,7 +68,7 @@ pub enum Bson {
     /// [ObjectId](http://dochub.mongodb.org/core/objectids)
     ObjectId(oid::ObjectId),
     /// UTC datetime
-    DateTime(chrono::DateTime<Utc>),
+    DateTime(crate::DateTime),
     /// Symbol (Deprecated)
     Symbol(String),
     /// [128-bit decimal floating point](https://github.com/mongodb/specifications/blob/master/source/bson-decimal128/decimal128.rst)
@@ -131,12 +130,12 @@ impl Display for Bson {
             }
             Bson::Binary(Binary { subtype, ref bytes }) => write!(
                 fmt,
-                "BinData({:#x}, {})",
+                "Binary({:#x}, {})",
                 u8::from(subtype),
                 base64::encode(bytes)
             ),
             Bson::ObjectId(ref id) => write!(fmt, "ObjectId(\"{}\")", id),
-            Bson::DateTime(date_time) => write!(fmt, "Date(\"{}\")", date_time),
+            Bson::DateTime(date_time) => write!(fmt, "DateTime(\"{}\")", date_time),
             Bson::Symbol(ref sym) => write!(fmt, "Symbol(\"{}\")", sym),
             Bson::Decimal128(ref d) => write!(fmt, "{}", d),
             Bson::Undefined => write!(fmt, "undefined"),
@@ -145,14 +144,48 @@ impl Display for Bson {
             Bson::DbPointer(DbPointer {
                 ref namespace,
                 ref id,
-            }) => write!(fmt, "DBPointer({}, {})", namespace, id),
+            }) => write!(fmt, "DbPointer({}, {})", namespace, id),
+        }
+    }
+}
+
+impl Debug for Bson {
+    fn fmt(&self, fmt: &mut Formatter) -> fmt::Result {
+        match *self {
+            Bson::Double(f) => fmt.debug_tuple("Double").field(&f).finish(),
+            Bson::String(ref s) => fmt.debug_tuple("String").field(s).finish(),
+            Bson::Array(ref vec) => {
+                write!(fmt, "Array(")?;
+                Debug::fmt(vec, fmt)?;
+                write!(fmt, ")")
+            }
+            Bson::Document(ref doc) => Debug::fmt(doc, fmt),
+            Bson::Boolean(b) => fmt.debug_tuple("Boolean").field(&b).finish(),
+            Bson::Null => write!(fmt, "Null"),
+            Bson::RegularExpression(ref regex) => Debug::fmt(regex, fmt),
+            Bson::JavaScriptCode(ref code) => {
+                fmt.debug_tuple("JavaScriptCode").field(code).finish()
+            }
+            Bson::JavaScriptCodeWithScope(ref code) => Debug::fmt(code, fmt),
+            Bson::Int32(i) => fmt.debug_tuple("Int32").field(&i).finish(),
+            Bson::Int64(i) => fmt.debug_tuple("Int64").field(&i).finish(),
+            Bson::Timestamp(ref t) => Debug::fmt(t, fmt),
+            Bson::Binary(ref b) => Debug::fmt(b, fmt),
+            Bson::ObjectId(ref id) => Debug::fmt(id, fmt),
+            Bson::DateTime(ref date_time) => Debug::fmt(date_time, fmt),
+            Bson::Symbol(ref sym) => fmt.debug_tuple("Symbol").field(sym).finish(),
+            Bson::Decimal128(ref d) => Debug::fmt(d, fmt),
+            Bson::Undefined => write!(fmt, "Undefined"),
+            Bson::MinKey => write!(fmt, "MinKey"),
+            Bson::MaxKey => write!(fmt, "MaxKey"),
+            Bson::DbPointer(ref pointer) => Debug::fmt(pointer, fmt),
         }
     }
 }
 
 impl From<f32> for Bson {
     fn from(a: f32) -> Bson {
-        Bson::Double(a as f64)
+        Bson::Double(a.into())
     }
 }
 
@@ -275,20 +308,9 @@ impl From<u32> for Bson {
     }
 }
 
-impl From<u64> for Bson {
-    /// This conversion is lossy if the provided `u64` is greater than `i64::MAX`, which it will be
-    /// converted to. Using this `From` implementation is highly discouraged and it will be
-    /// removed in the next major version.
-    ///
-    /// Note: due to https://github.com/rust-lang/rust/issues/39935 we cannot deprecate this implementation.
-    fn from(a: u64) -> Bson {
-        Bson::Int64(a.try_into().unwrap_or(i64::MAX))
-    }
-}
-
 impl From<[u8; 12]> for Bson {
     fn from(a: [u8; 12]) -> Bson {
-        Bson::ObjectId(oid::ObjectId::with_bytes(a))
+        Bson::ObjectId(oid::ObjectId::from_bytes(a))
     }
 }
 
@@ -298,9 +320,25 @@ impl From<oid::ObjectId> for Bson {
     }
 }
 
-impl From<chrono::DateTime<Utc>> for Bson {
-    fn from(a: chrono::DateTime<Utc>) -> Bson {
-        Bson::DateTime(a)
+#[cfg(feature = "chrono-0_4")]
+#[cfg_attr(docsrs, doc(cfg(feature = "chrono-0_4")))]
+impl<T: chrono::TimeZone> From<chrono::DateTime<T>> for Bson {
+    fn from(a: chrono::DateTime<T>) -> Bson {
+        Bson::DateTime(crate::DateTime::from(a))
+    }
+}
+
+#[cfg(feature = "uuid-0_8")]
+#[cfg_attr(docsrs, doc(cfg(feature = "uuid-0_8")))]
+impl From<uuid::Uuid> for Bson {
+    fn from(uuid: uuid::Uuid) -> Self {
+        Bson::Binary(uuid.into())
+    }
+}
+
+impl From<crate::DateTime> for Bson {
+    fn from(dt: crate::DateTime) -> Self {
+        Bson::DateTime(dt)
     }
 }
 
@@ -310,6 +348,18 @@ impl From<DbPointer> for Bson {
     }
 }
 
+impl<T> From<Option<T>> for Bson
+where
+    T: Into<Bson>,
+{
+    fn from(a: Option<T>) -> Bson {
+        match a {
+            None => Bson::Null,
+            Some(t) => t.into(),
+        }
+    }
+}
+
 /// This will create the [relaxed Extended JSON v2](https://docs.mongodb.com/manual/reference/mongodb-extended-json/) representation of the provided [`Bson`](../enum.Bson.html).
 impl From<Bson> for Value {
     fn from(bson: Bson) -> Self {
@@ -320,9 +370,7 @@ impl From<Bson> for Value {
 impl Bson {
     /// Converts the Bson value into its [relaxed extended JSON representation](https://docs.mongodb.com/manual/reference/mongodb-extended-json/).
     ///
-    /// Note: extended json encoding for `Decimal128` values is not supported without the
-    /// "decimal128" feature flag. If this method is called on a case which contains a
-    /// `Decimal128` value, it will panic.
+    /// Note: If this method is called on a case which contains a `Decimal128` value, it will panic.
     pub fn into_relaxed_extjson(self) -> Value {
         match self {
             Bson::Double(v) if v.is_nan() => {
@@ -383,28 +431,16 @@ impl Bson {
                 })
             }
             Bson::ObjectId(v) => json!({"$oid": v.to_hex()}),
-            Bson::DateTime(v) if v.timestamp_millis() >= 0 && v.year() <= 99999 => {
-                let seconds_format = if v.timestamp_subsec_millis() == 0 {
-                    SecondsFormat::Secs
-                } else {
-                    SecondsFormat::Millis
-                };
-
+            Bson::DateTime(v) if v.timestamp_millis() >= 0 && v.to_chrono().year() <= 99999 => {
                 json!({
-                    "$date": v.to_rfc3339_opts(seconds_format, true),
+                    "$date": v.to_rfc3339(),
                 })
             }
             Bson::DateTime(v) => json!({
                 "$date": { "$numberLong": v.timestamp_millis().to_string() },
             }),
             Bson::Symbol(v) => json!({ "$symbol": v }),
-            #[cfg(feature = "decimal128")]
-            Bson::Decimal128(ref v) => json!({ "$numberDecimal": v.to_string() }),
-            #[cfg(not(feature = "decimal128"))]
-            Bson::Decimal128(_) => panic!(
-                "Decimal128 extended JSON not implemented yet. Use the decimal128 feature to \
-                 enable experimental support for it."
-            ),
+            Bson::Decimal128(_) => panic!("Decimal128 extended JSON not implemented yet."),
             Bson::Undefined => json!({ "$undefined": true }),
             Bson::MinKey => json!({ "$minKey": 1 }),
             Bson::MaxKey => json!({ "$maxKey": 1 }),
@@ -424,9 +460,8 @@ impl Bson {
 
     /// Converts the Bson value into its [canonical extended JSON representation](https://docs.mongodb.com/manual/reference/mongodb-extended-json/).
     ///
-    /// Note: extended json encoding for `Decimal128` values is not supported without the
-    /// "decimal128" feature flag. If this method is called on a case which contains a
-    /// `Decimal128` value, it will panic.
+    /// Note: extended json encoding for `Decimal128` values is not supported. If this method is
+    /// called on a case which contains a `Decimal128` value, it will panic.
     pub fn into_canonical_extjson(self) -> Value {
         match self {
             Bson::Int32(i) => json!({ "$numberInt": i.to_string() }),
@@ -495,8 +530,8 @@ impl Bson {
     /// This function mainly used for [extended JSON format](https://docs.mongodb.com/manual/reference/mongodb-extended-json/).
     // TODO RUST-426: Investigate either removing this from the serde implementation or unifying
     // with the extended JSON implementation.
-    pub(crate) fn to_extended_document(&self) -> Document {
-        match *self {
+    pub(crate) fn into_extended_document(self) -> Document {
+        match self {
             Bson::RegularExpression(Regex {
                 ref pattern,
                 ref options,
@@ -508,23 +543,20 @@ impl Bson {
 
                 doc! {
                     "$regularExpression": {
-                        "pattern": pattern.clone(),
+                        "pattern": pattern,
                         "options": options,
                     }
                 }
             }
             Bson::JavaScriptCode(ref code) => {
                 doc! {
-                    "$code": code.clone(),
+                    "$code": code,
                 }
             }
-            Bson::JavaScriptCodeWithScope(JavaScriptCodeWithScope {
-                ref code,
-                ref scope,
-            }) => {
+            Bson::JavaScriptCodeWithScope(JavaScriptCodeWithScope { code, scope }) => {
                 doc! {
-                    "$code": code.clone(),
-                    "$scope": scope.clone(),
+                    "$code": code,
+                    "$scope": scope,
                 }
             }
             Bson::Timestamp(Timestamp { time, increment }) => {
@@ -549,15 +581,9 @@ impl Bson {
                     "$oid": v.to_string(),
                 }
             }
-            Bson::DateTime(v) if v.timestamp_millis() >= 0 && v.year() <= 99999 => {
-                let seconds_format = if v.timestamp_subsec_millis() == 0 {
-                    SecondsFormat::Secs
-                } else {
-                    SecondsFormat::Millis
-                };
-
+            Bson::DateTime(v) if v.timestamp_millis() >= 0 && v.to_chrono().year() <= 9999 => {
                 doc! {
-                    "$date": v.to_rfc3339_opts(seconds_format, true),
+                    "$date": v.to_rfc3339(),
                 }
             }
             Bson::DateTime(v) => doc! {
@@ -568,21 +594,6 @@ impl Bson {
                     "$symbol": v.to_owned(),
                 }
             }
-            #[cfg(feature = "decimal128")]
-            Bson::Decimal128(ref v) => {
-                doc! {
-                    "$numberDecimal": (v.to_string())
-                }
-            }
-            #[cfg(not(feature = "decimal128"))]
-            Bson::Decimal128(ref v) => {
-                doc! {
-                    "$numberDecimalBytes": Bson::Binary(Binary {
-                        bytes: v.bytes.to_vec(),
-                        subtype: BinarySubtype::Generic,
-                    }).to_extended_document()
-                }
-            }
             Bson::Undefined => {
                 doc! {
                     "$undefined": true,
@@ -626,7 +637,7 @@ impl Bson {
         match keys.as_slice() {
             ["$oid"] => {
                 if let Ok(oid) = doc.get_str("$oid") {
-                    if let Ok(oid) = ObjectId::with_string(oid) {
+                    if let Ok(oid) = ObjectId::parse_str(oid) {
                         return Bson::ObjectId(oid);
                     }
                 }
@@ -666,21 +677,10 @@ impl Bson {
                 _ => {}
             },
 
-            #[cfg(feature = "decimal128")]
-            ["$numberDecimal"] => {
-                if let Ok(d) = doc.get_str("$numberDecimal") {
-                    if let Ok(d) = d.parse() {
-                        return Bson::Decimal128(d);
-                    }
-                }
-            }
-
-            #[cfg(not(feature = "decimal128"))]
             ["$numberDecimalBytes"] => {
-                if let Ok(d) = doc.get_binary_generic("$numberDecimalBytes") {
-                    let boxed_slice = d.clone().into_boxed_slice();
-                    if let Ok(ba) = Box::<[u8; 128 / 8]>::try_from(boxed_slice) {
-                        return Bson::Decimal128(Decimal128 { bytes: *ba });
+                if let Ok(bytes) = doc.get_binary_generic("$numberDecimalBytes") {
+                    if let Ok(b) = bytes.clone().try_into() {
+                        return Bson::Decimal128(Decimal128 { bytes: b });
                     }
                 }
             }
@@ -758,7 +758,7 @@ impl Bson {
                         if let Ok(id) = db_pointer.get_object_id("$id") {
                             return Bson::DbPointer(DbPointer {
                                 namespace: ns.into(),
-                                id: id.clone(),
+                                id,
                             });
                         }
                     }
@@ -767,12 +767,12 @@ impl Bson {
 
             ["$date"] => {
                 if let Ok(date) = doc.get_i64("$date") {
-                    return Bson::DateTime(DateTime::from_i64(date).into());
+                    return Bson::DateTime(crate::DateTime::from_millis(date));
                 }
 
                 if let Ok(date) = doc.get_str("$date") {
                     if let Ok(date) = chrono::DateTime::parse_from_rfc3339(date) {
-                        return Bson::DateTime(date.into());
+                        return Bson::DateTime(crate::DateTime::from_chrono(date));
                     }
                 }
             }
@@ -901,9 +901,9 @@ impl Bson {
     }
 
     /// If `Bson` is `Objectid`, return its value. Returns `None` otherwise
-    pub fn as_object_id(&self) -> Option<&oid::ObjectId> {
+    pub fn as_object_id(&self) -> Option<oid::ObjectId> {
         match *self {
-            Bson::ObjectId(ref v) => Some(v),
+            Bson::ObjectId(v) => Some(v),
             _ => None,
         }
     }
@@ -917,7 +917,7 @@ impl Bson {
     }
 
     /// If `Bson` is `DateTime`, return its value. Returns `None` otherwise
-    pub fn as_datetime(&self) -> Option<&chrono::DateTime<Utc>> {
+    pub fn as_datetime(&self) -> Option<&crate::DateTime> {
         match *self {
             Bson::DateTime(ref v) => Some(v),
             _ => None,
@@ -926,7 +926,7 @@ impl Bson {
 
     /// If `Bson` is `DateTime`, return a mutable reference to its value. Returns `None`
     /// otherwise
-    pub fn as_datetime_mut(&mut self) -> Option<&mut chrono::DateTime<Utc>> {
+    pub fn as_datetime_mut(&mut self) -> Option<&mut crate::DateTime> {
         match *self {
             Bson::DateTime(ref mut v) => Some(v),
             _ => None,
@@ -1002,79 +1002,6 @@ impl Timestamp {
     }
 }
 
-/// `DateTime` representation in struct for serde serialization
-///
-/// Just a helper for convenience
-///
-/// ```rust
-/// use serde::{Serialize, Deserialize};
-/// use bson::DateTime;
-///
-/// #[derive(Serialize, Deserialize)]
-/// struct Foo {
-///     date_time: DateTime,
-/// }
-/// ```
-#[derive(Debug, Eq, PartialEq, Ord, PartialOrd, Hash, Copy, Clone)]
-pub struct DateTime(pub chrono::DateTime<Utc>);
-
-impl DateTime {
-    pub(crate) fn from_i64(date: i64) -> Self {
-        let mut num_secs = date / 1000;
-        let mut num_millis = date % 1000;
-
-        // The chrono API only lets us create a DateTime with an i64 number of seconds
-        // and a u32 number of nanoseconds. In the case of a negative timestamp, this
-        // means that we need to turn the negative fractional part into a positive and
-        // shift the number of seconds down. For example:
-        //
-        //     date       = -4300 ms
-        //     num_secs   = date / 1000 = -4300 / 1000 = -4
-        //     num_millis = date % 1000 = -4300 % 1000 = -300
-        //
-        // Since num_millis is less than 0:
-        //     num_secs   = num_secs -1 = -4 - 1 = -5
-        //     num_millis = num_nanos + 1000 = -300 + 1000 = 700
-        //
-        // Instead of -4 seconds and -300 milliseconds, we now have -5 seconds and +700
-        // milliseconds, which expresses the same timestamp, but in a way we can create
-        // a DateTime with.
-        if num_millis < 0 {
-            num_secs -= 1;
-            num_millis += 1000;
-        };
-
-        Utc.timestamp(num_secs, num_millis as u32 * 1_000_000)
-            .into()
-    }
-}
-
-impl Deref for DateTime {
-    type Target = chrono::DateTime<Utc>;
-
-    fn deref(&self) -> &Self::Target {
-        &self.0
-    }
-}
-
-impl DerefMut for DateTime {
-    fn deref_mut(&mut self) -> &mut chrono::DateTime<Utc> {
-        &mut self.0
-    }
-}
-
-impl From<DateTime> for chrono::DateTime<Utc> {
-    fn from(utc: DateTime) -> Self {
-        utc.0
-    }
-}
-
-impl From<chrono::DateTime<Utc>> for DateTime {
-    fn from(x: chrono::DateTime<Utc>) -> Self {
-        DateTime(x)
-    }
-}
-
 /// Represents a BSON regular expression value.
 #[derive(Debug, Clone, PartialEq)]
 pub struct Regex {
@@ -1127,6 +1054,17 @@ impl Binary {
     }
 }
 
+#[cfg(feature = "uuid-0_8")]
+#[cfg_attr(docsrs, doc(cfg(feature = "uuid-0_8")))]
+impl From<uuid::Uuid> for Binary {
+    fn from(uuid: uuid::Uuid) -> Self {
+        Binary {
+            subtype: BinarySubtype::Uuid,
+            bytes: uuid.as_bytes().to_vec(),
+        }
+    }
+}
+
 /// Represents a DBPointer. (Deprecated)
 #[derive(Debug, Clone, PartialEq)]
 pub struct DbPointer {
