@@ -4,30 +4,26 @@ use std::{
     error,
     fmt::{self, Debug, Display, Formatter},
     io::{Read, Write},
-    iter::{Extend, FromIterator, IntoIterator, Map},
-    marker::PhantomData,
+    iter::{Extend, FromIterator, IntoIterator},
     mem,
 };
 
-use chrono::{DateTime, Utc};
+use ahash::RandomState;
+use indexmap::IndexMap;
+use serde::de::Error;
 
-use linked_hash_map::{self, LinkedHashMap};
-
-use serde::de::{self, Error, MapAccess, Visitor};
-
-#[cfg(feature = "decimal128")]
-use crate::decimal128::Decimal128;
 use crate::{
     bson::{Array, Binary, Bson, Timestamp},
     de::{deserialize_bson_kvp, ensure_read_exactly, read_i32, MIN_BSON_DOCUMENT_SIZE},
     oid::ObjectId,
     ser::{serialize_bson, write_i32},
     spec::BinarySubtype,
+    Decimal128,
 };
 
 /// Error to indicate that either a value was empty or it contained an unexpected
 /// type, for use with the direct getters.
-#[derive(PartialEq)]
+#[derive(PartialEq, Clone)]
 #[non_exhaustive]
 pub enum ValueAccessError {
     /// Cannot find the expected field with the specified key
@@ -64,7 +60,7 @@ impl error::Error for ValueAccessError {}
 /// A BSON document represented as an associative HashMap with insertion ordering.
 #[derive(Clone, PartialEq)]
 pub struct Document {
-    inner: LinkedHashMap<String, Bson>,
+    inner: IndexMap<String, Bson, RandomState>,
 }
 
 impl Default for Document {
@@ -86,7 +82,7 @@ impl Display for Document {
                 fmt.write_str(", ")?;
             }
 
-            write!(fmt, "{}: {}", k, v)?;
+            write!(fmt, "\"{}\": {}", k, v)?;
         }
 
         write!(fmt, "{}}}", if !first { " " } else { "" })
@@ -94,31 +90,31 @@ impl Display for Document {
 }
 
 impl Debug for Document {
-    fn fmt(&self, f: &mut Formatter) -> fmt::Result {
-        write!(f, "Document({:?})", self.inner)
+    fn fmt(&self, fmt: &mut Formatter) -> fmt::Result {
+        write!(fmt, "Document(")?;
+        Debug::fmt(&self.inner, fmt)?;
+        write!(fmt, ")")
     }
 }
 
 /// An iterator over Document entries.
-pub struct DocumentIntoIterator {
-    inner: LinkedHashMap<String, Bson>,
+pub struct IntoIter {
+    inner: indexmap::map::IntoIter<String, Bson>,
 }
 
 /// An owning iterator over Document entries.
-pub struct DocumentIterator<'a> {
-    inner: linked_hash_map::Iter<'a, String, Bson>,
+pub struct Iter<'a> {
+    inner: indexmap::map::Iter<'a, String, Bson>,
 }
 
-type DocumentMap<'a, T> = Map<DocumentIterator<'a>, fn((&'a String, &'a Bson)) -> T>;
-
 /// An iterator over an Document's keys.
 pub struct Keys<'a> {
-    inner: DocumentMap<'a, &'a String>,
+    inner: indexmap::map::Keys<'a, String, Bson>,
 }
 
 /// An iterator over an Document's values.
 pub struct Values<'a> {
-    inner: DocumentMap<'a, &'a Bson>,
+    inner: indexmap::map::Values<'a, String, Bson>,
 }
 
 impl<'a> Iterator for Keys<'a> {
@@ -139,19 +135,21 @@ impl<'a> Iterator for Values<'a> {
 
 impl IntoIterator for Document {
     type Item = (String, Bson);
-    type IntoIter = DocumentIntoIterator;
+    type IntoIter = IntoIter;
 
     fn into_iter(self) -> Self::IntoIter {
-        DocumentIntoIterator { inner: self.inner }
+        IntoIter {
+            inner: self.inner.into_iter(),
+        }
     }
 }
 
 impl<'a> IntoIterator for &'a Document {
     type Item = (&'a String, &'a Bson);
-    type IntoIter = DocumentIterator<'a>;
+    type IntoIter = Iter<'a>;
 
     fn into_iter(self) -> Self::IntoIter {
-        DocumentIterator {
+        Iter {
             inner: self.inner.iter(),
         }
     }
@@ -167,15 +165,15 @@ impl FromIterator<(String, Bson)> for Document {
     }
 }
 
-impl<'a> Iterator for DocumentIntoIterator {
+impl<'a> Iterator for IntoIter {
     type Item = (String, Bson);
 
     fn next(&mut self) -> Option<(String, Bson)> {
-        self.inner.pop_front()
+        self.inner.next()
     }
 }
 
-impl<'a> Iterator for DocumentIterator<'a> {
+impl<'a> Iterator for Iter<'a> {
     type Item = (&'a String, &'a Bson);
 
     fn next(&mut self) -> Option<(&'a String, &'a Bson)> {
@@ -187,12 +185,12 @@ impl Document {
     /// Creates a new empty Document.
     pub fn new() -> Document {
         Document {
-            inner: LinkedHashMap::new(),
+            inner: IndexMap::default(),
         }
     }
 
     /// Gets an iterator over the entries of the map.
-    pub fn iter(&self) -> DocumentIterator {
+    pub fn iter(&self) -> Iter {
         self.into_iter()
     }
 
@@ -202,18 +200,18 @@ impl Document {
     }
 
     /// Returns a reference to the Bson corresponding to the key.
-    pub fn get(&self, key: &str) -> Option<&Bson> {
-        self.inner.get(key)
+    pub fn get(&self, key: impl AsRef<str>) -> Option<&Bson> {
+        self.inner.get(key.as_ref())
     }
 
     /// Gets a mutable reference to the Bson corresponding to the key
-    pub fn get_mut(&mut self, key: &str) -> Option<&mut Bson> {
-        self.inner.get_mut(key)
+    pub fn get_mut(&mut self, key: impl AsRef<str>) -> Option<&mut Bson> {
+        self.inner.get_mut(key.as_ref())
     }
 
     /// Get a floating point value for this key if it exists and has
     /// the correct type.
-    pub fn get_f64(&self, key: &str) -> ValueAccessResult<f64> {
+    pub fn get_f64(&self, key: impl AsRef<str>) -> ValueAccessResult<f64> {
         match self.get(key) {
             Some(&Bson::Double(v)) => Ok(v),
             Some(_) => Err(ValueAccessError::UnexpectedType),
@@ -223,7 +221,7 @@ impl Document {
 
     /// Get a mutable reference to a floating point value for this key if it exists and has
     /// the correct type.
-    pub fn get_f64_mut(&mut self, key: &str) -> ValueAccessResult<&mut f64> {
+    pub fn get_f64_mut(&mut self, key: impl AsRef<str>) -> ValueAccessResult<&mut f64> {
         match self.get_mut(key) {
             Some(&mut Bson::Double(ref mut v)) => Ok(v),
             Some(_) => Err(ValueAccessError::UnexpectedType),
@@ -232,8 +230,7 @@ impl Document {
     }
 
     /// Get a reference to a Decimal128 value for key, if it exists.
-    #[cfg(feature = "decimal128")]
-    pub fn get_decimal128(&self, key: &str) -> ValueAccessResult<&Decimal128> {
+    pub fn get_decimal128(&self, key: impl AsRef<str>) -> ValueAccessResult<&Decimal128> {
         match self.get(key) {
             Some(&Bson::Decimal128(ref v)) => Ok(v),
             Some(_) => Err(ValueAccessError::UnexpectedType),
@@ -242,8 +239,10 @@ impl Document {
     }
 
     /// Get a mutable reference to a Decimal128 value for key, if it exists.
-    #[cfg(feature = "decimal128")]
-    pub fn get_decimal128_mut(&mut self, key: &str) -> ValueAccessResult<&mut Decimal128> {
+    pub fn get_decimal128_mut(
+        &mut self,
+        key: impl AsRef<str>,
+    ) -> ValueAccessResult<&mut Decimal128> {
         match self.get_mut(key) {
             Some(&mut Bson::Decimal128(ref mut v)) => Ok(v),
             Some(_) => Err(ValueAccessError::UnexpectedType),
@@ -252,7 +251,7 @@ impl Document {
     }
 
     /// Get a string slice this key if it exists and has the correct type.
-    pub fn get_str(&self, key: &str) -> ValueAccessResult<&str> {
+    pub fn get_str(&self, key: impl AsRef<str>) -> ValueAccessResult<&str> {
         match self.get(key) {
             Some(&Bson::String(ref v)) => Ok(v),
             Some(_) => Err(ValueAccessError::UnexpectedType),
@@ -261,7 +260,7 @@ impl Document {
     }
 
     /// Get a mutable string slice this key if it exists and has the correct type.
-    pub fn get_str_mut(&mut self, key: &str) -> ValueAccessResult<&mut str> {
+    pub fn get_str_mut(&mut self, key: impl AsRef<str>) -> ValueAccessResult<&mut str> {
         match self.get_mut(key) {
             Some(&mut Bson::String(ref mut v)) => Ok(v),
             Some(_) => Err(ValueAccessError::UnexpectedType),
@@ -271,7 +270,7 @@ impl Document {
 
     /// Get a reference to an array for this key if it exists and has
     /// the correct type.
-    pub fn get_array(&self, key: &str) -> ValueAccessResult<&Array> {
+    pub fn get_array(&self, key: impl AsRef<str>) -> ValueAccessResult<&Array> {
         match self.get(key) {
             Some(&Bson::Array(ref v)) => Ok(v),
             Some(_) => Err(ValueAccessError::UnexpectedType),
@@ -281,7 +280,7 @@ impl Document {
 
     /// Get a mutable reference to an array for this key if it exists and has
     /// the correct type.
-    pub fn get_array_mut(&mut self, key: &str) -> ValueAccessResult<&mut Array> {
+    pub fn get_array_mut(&mut self, key: impl AsRef<str>) -> ValueAccessResult<&mut Array> {
         match self.get_mut(key) {
             Some(&mut Bson::Array(ref mut v)) => Ok(v),
             Some(_) => Err(ValueAccessError::UnexpectedType),
@@ -291,7 +290,7 @@ impl Document {
 
     /// Get a reference to a document for this key if it exists and has
     /// the correct type.
-    pub fn get_document(&self, key: &str) -> ValueAccessResult<&Document> {
+    pub fn get_document(&self, key: impl AsRef<str>) -> ValueAccessResult<&Document> {
         match self.get(key) {
             Some(&Bson::Document(ref v)) => Ok(v),
             Some(_) => Err(ValueAccessError::UnexpectedType),
@@ -301,7 +300,7 @@ impl Document {
 
     /// Get a mutable reference to a document for this key if it exists and has
     /// the correct type.
-    pub fn get_document_mut(&mut self, key: &str) -> ValueAccessResult<&mut Document> {
+    pub fn get_document_mut(&mut self, key: impl AsRef<str>) -> ValueAccessResult<&mut Document> {
         match self.get_mut(key) {
             Some(&mut Bson::Document(ref mut v)) => Ok(v),
             Some(_) => Err(ValueAccessError::UnexpectedType),
@@ -310,7 +309,7 @@ impl Document {
     }
 
     /// Get a bool value for this key if it exists and has the correct type.
-    pub fn get_bool(&self, key: &str) -> ValueAccessResult<bool> {
+    pub fn get_bool(&self, key: impl AsRef<str>) -> ValueAccessResult<bool> {
         match self.get(key) {
             Some(&Bson::Boolean(v)) => Ok(v),
             Some(_) => Err(ValueAccessError::UnexpectedType),
@@ -319,7 +318,7 @@ impl Document {
     }
 
     /// Get a mutable reference to a bool value for this key if it exists and has the correct type.
-    pub fn get_bool_mut(&mut self, key: &str) -> ValueAccessResult<&mut bool> {
+    pub fn get_bool_mut(&mut self, key: impl AsRef<str>) -> ValueAccessResult<&mut bool> {
         match self.get_mut(key) {
             Some(&mut Bson::Boolean(ref mut v)) => Ok(v),
             Some(_) => Err(ValueAccessError::UnexpectedType),
@@ -328,12 +327,12 @@ impl Document {
     }
 
     /// Returns wether this key has a null value
-    pub fn is_null(&self, key: &str) -> bool {
+    pub fn is_null(&self, key: impl AsRef<str>) -> bool {
         self.get(key) == Some(&Bson::Null)
     }
 
     /// Get an i32 value for this key if it exists and has the correct type.
-    pub fn get_i32(&self, key: &str) -> ValueAccessResult<i32> {
+    pub fn get_i32(&self, key: impl AsRef<str>) -> ValueAccessResult<i32> {
         match self.get(key) {
             Some(&Bson::Int32(v)) => Ok(v),
             Some(_) => Err(ValueAccessError::UnexpectedType),
@@ -342,7 +341,7 @@ impl Document {
     }
 
     /// Get a mutable reference to an i32 value for this key if it exists and has the correct type.
-    pub fn get_i32_mut(&mut self, key: &str) -> ValueAccessResult<&mut i32> {
+    pub fn get_i32_mut(&mut self, key: impl AsRef<str>) -> ValueAccessResult<&mut i32> {
         match self.get_mut(key) {
             Some(&mut Bson::Int32(ref mut v)) => Ok(v),
             Some(_) => Err(ValueAccessError::UnexpectedType),
@@ -351,7 +350,7 @@ impl Document {
     }
 
     /// Get an i64 value for this key if it exists and has the correct type.
-    pub fn get_i64(&self, key: &str) -> ValueAccessResult<i64> {
+    pub fn get_i64(&self, key: impl AsRef<str>) -> ValueAccessResult<i64> {
         match self.get(key) {
             Some(&Bson::Int64(v)) => Ok(v),
             Some(_) => Err(ValueAccessError::UnexpectedType),
@@ -360,7 +359,7 @@ impl Document {
     }
 
     /// Get a mutable reference to an i64 value for this key if it exists and has the correct type.
-    pub fn get_i64_mut(&mut self, key: &str) -> ValueAccessResult<&mut i64> {
+    pub fn get_i64_mut(&mut self, key: impl AsRef<str>) -> ValueAccessResult<&mut i64> {
         match self.get_mut(key) {
             Some(&mut Bson::Int64(ref mut v)) => Ok(v),
             Some(_) => Err(ValueAccessError::UnexpectedType),
@@ -369,7 +368,7 @@ impl Document {
     }
 
     /// Get a time stamp value for this key if it exists and has the correct type.
-    pub fn get_timestamp(&self, key: &str) -> ValueAccessResult<Timestamp> {
+    pub fn get_timestamp(&self, key: impl AsRef<str>) -> ValueAccessResult<Timestamp> {
         match self.get(key) {
             Some(&Bson::Timestamp(timestamp)) => Ok(timestamp),
             Some(_) => Err(ValueAccessError::UnexpectedType),
@@ -379,7 +378,7 @@ impl Document {
 
     /// Get a mutable reference to a time stamp value for this key if it exists and has the correct
     /// type.
-    pub fn get_timestamp_mut(&mut self, key: &str) -> ValueAccessResult<&mut Timestamp> {
+    pub fn get_timestamp_mut(&mut self, key: impl AsRef<str>) -> ValueAccessResult<&mut Timestamp> {
         match self.get_mut(key) {
             Some(&mut Bson::Timestamp(ref mut timestamp)) => Ok(timestamp),
             Some(_) => Err(ValueAccessError::UnexpectedType),
@@ -389,7 +388,7 @@ impl Document {
 
     /// Get a reference to a generic binary value for this key if it exists and has the correct
     /// type.
-    pub fn get_binary_generic(&self, key: &str) -> ValueAccessResult<&Vec<u8>> {
+    pub fn get_binary_generic(&self, key: impl AsRef<str>) -> ValueAccessResult<&Vec<u8>> {
         match self.get(key) {
             Some(&Bson::Binary(Binary {
                 subtype: BinarySubtype::Generic,
@@ -402,7 +401,10 @@ impl Document {
 
     /// Get a mutable reference generic binary value for this key if it exists and has the correct
     /// type.
-    pub fn get_binary_generic_mut(&mut self, key: &str) -> ValueAccessResult<&mut Vec<u8>> {
+    pub fn get_binary_generic_mut(
+        &mut self,
+        key: impl AsRef<str>,
+    ) -> ValueAccessResult<&mut Vec<u8>> {
         match self.get_mut(key) {
             Some(&mut Bson::Binary(Binary {
                 subtype: BinarySubtype::Generic,
@@ -413,10 +415,10 @@ impl Document {
         }
     }
 
-    /// Get a reference to an object id value for this key if it exists and has the correct type.
-    pub fn get_object_id(&self, key: &str) -> ValueAccessResult<&ObjectId> {
+    /// Get an object id value for this key if it exists and has the correct type.
+    pub fn get_object_id(&self, key: impl AsRef<str>) -> ValueAccessResult<ObjectId> {
         match self.get(key) {
-            Some(&Bson::ObjectId(ref v)) => Ok(v),
+            Some(&Bson::ObjectId(v)) => Ok(v),
             Some(_) => Err(ValueAccessError::UnexpectedType),
             None => Err(ValueAccessError::NotPresent),
         }
@@ -424,7 +426,7 @@ impl Document {
 
     /// Get a mutable reference to an object id value for this key if it exists and has the correct
     /// type.
-    pub fn get_object_id_mut(&mut self, key: &str) -> ValueAccessResult<&mut ObjectId> {
+    pub fn get_object_id_mut(&mut self, key: impl AsRef<str>) -> ValueAccessResult<&mut ObjectId> {
         match self.get_mut(key) {
             Some(&mut Bson::ObjectId(ref mut v)) => Ok(v),
             Some(_) => Err(ValueAccessError::UnexpectedType),
@@ -433,7 +435,7 @@ impl Document {
     }
 
     /// Get a reference to a UTC datetime value for this key if it exists and has the correct type.
-    pub fn get_datetime(&self, key: &str) -> ValueAccessResult<&DateTime<Utc>> {
+    pub fn get_datetime(&self, key: impl AsRef<str>) -> ValueAccessResult<&crate::DateTime> {
         match self.get(key) {
             Some(&Bson::DateTime(ref v)) => Ok(v),
             Some(_) => Err(ValueAccessError::UnexpectedType),
@@ -443,7 +445,10 @@ impl Document {
 
     /// Get a mutable reference to a UTC datetime value for this key if it exists and has the
     /// correct type.
-    pub fn get_datetime_mut(&mut self, key: &str) -> ValueAccessResult<&mut DateTime<Utc>> {
+    pub fn get_datetime_mut(
+        &mut self,
+        key: impl AsRef<str>,
+    ) -> ValueAccessResult<&mut crate::DateTime> {
         match self.get_mut(key) {
             Some(&mut Bson::DateTime(ref mut v)) => Ok(v),
             Some(_) => Err(ValueAccessError::UnexpectedType),
@@ -452,31 +457,21 @@ impl Document {
     }
 
     /// Returns true if the map contains a value for the specified key.
-    pub fn contains_key(&self, key: &str) -> bool {
-        self.inner.contains_key(key)
+    pub fn contains_key(&self, key: impl AsRef<str>) -> bool {
+        self.inner.contains_key(key.as_ref())
     }
 
     /// Gets a collection of all keys in the document.
-    pub fn keys<'a>(&'a self) -> Keys<'a> {
-        fn first<A, B>((a, _): (A, B)) -> A {
-            a
-        }
-        let first: fn((&'a String, &'a Bson)) -> &'a String = first;
-
+    pub fn keys(&self) -> Keys {
         Keys {
-            inner: self.iter().map(first),
+            inner: self.inner.keys(),
         }
     }
 
     /// Gets a collection of all values in the document.
-    pub fn values<'a>(&'a self) -> Values<'a> {
-        fn second<A, B>((_, b): (A, B)) -> B {
-            b
-        }
-        let second: fn((&'a String, &'a Bson)) -> &'a Bson = second;
-
+    pub fn values(&self) -> Values {
         Values {
-            inner: self.iter().map(second),
+            inner: self.inner.values(),
         }
     }
 
@@ -498,25 +493,42 @@ impl Document {
     }
 
     /// Takes the value of the entry out of the document, and returns it.
-    pub fn remove(&mut self, key: &str) -> Option<Bson> {
-        self.inner.remove(key)
+    /// Computes in **O(n)** time (average).
+    pub fn remove(&mut self, key: impl AsRef<str>) -> Option<Bson> {
+        self.inner.shift_remove(key.as_ref())
     }
 
     pub fn entry(&mut self, k: String) -> Entry {
-        Entry {
-            inner: self.inner.entry(k),
+        match self.inner.entry(k) {
+            indexmap::map::Entry::Occupied(o) => Entry::Occupied(OccupiedEntry { inner: o }),
+            indexmap::map::Entry::Vacant(v) => Entry::Vacant(VacantEntry { inner: v }),
         }
     }
 
     /// Attempts to serialize the `Document` into a byte stream.
-    pub fn to_writer<W: Write + ?Sized>(&self, writer: &mut W) -> crate::ser::Result<()> {
+    ///
+    /// While the method signature indicates an owned writer must be passed in, a mutable reference
+    /// may also be passed in due to blanket implementations of `Write` provided in the standard
+    /// library.
+    ///
+    /// ```
+    /// # fn main() -> bson::ser::Result<()> {
+    /// use bson::doc;
+    ///
+    /// let mut v: Vec<u8> = Vec::new();
+    /// let doc = doc! { "x" : 1 };
+    /// doc.to_writer(&mut v)?;
+    /// # Ok(())
+    /// # }
+    /// ```
+    pub fn to_writer<W: Write>(&self, mut writer: W) -> crate::ser::Result<()> {
         let mut buf = Vec::new();
         for (key, val) in self.into_iter() {
             serialize_bson(&mut buf, key.as_ref(), val)?;
         }
 
         write_i32(
-            writer,
+            &mut writer,
             (buf.len() + mem::size_of::<i32>() + mem::size_of::<u8>()) as i32,
         )?;
         writer.write_all(&buf)?;
@@ -560,88 +572,111 @@ impl Document {
     }
 
     /// Attempts to deserialize a `Document` from a byte stream.
-    pub fn from_reader<R: Read + ?Sized>(reader: &mut R) -> crate::de::Result<Document> {
-        Self::decode(reader, false)
+    ///
+    /// While the method signature indicates an owned reader must be passed in, a mutable reference
+    /// may also be passed in due to blanket implementations of `Read` provided in the standard
+    /// library.
+    ///
+    /// ```
+    /// # use std::error::Error;
+    /// # fn main() -> std::result::Result<(), Box<dyn Error>> {
+    /// use bson::{doc, Document};
+    /// use std::io::Cursor;
+    ///
+    /// let mut v: Vec<u8> = Vec::new();
+    /// let doc = doc! { "x" : 1 };
+    /// doc.to_writer(&mut v)?;
+    ///
+    /// // read from mutable reference
+    /// let mut reader = Cursor::new(v.clone());
+    /// let doc1 = Document::from_reader(&mut reader)?;
+    ///
+    /// // read from owned value
+    /// let doc2 = Document::from_reader(Cursor::new(v))?;
+    ///
+    /// assert_eq!(doc, doc1);
+    /// assert_eq!(doc, doc2);
+    /// # Ok(())
+    /// # }
+    /// ```
+    pub fn from_reader<R: Read>(mut reader: R) -> crate::de::Result<Document> {
+        Self::decode(&mut reader, false)
     }
 
     /// Attempt to deserialize a `Document` that may contain invalid UTF-8 strings from a byte
     /// stream.
     ///
     /// This is mainly useful when reading raw BSON returned from a MongoDB server, which
-    /// in rare cases can contain invalidly truncated strings (https://jira.mongodb.org/browse/SERVER-24007).
+    /// in rare cases can contain invalidly truncated strings (<https://jira.mongodb.org/browse/SERVER-24007>).
     /// For most use cases, `Document::from_reader` can be used instead.
-    pub fn from_reader_utf8_lossy<R: Read + ?Sized>(reader: &mut R) -> crate::de::Result<Document> {
-        Self::decode(reader, true)
+    pub fn from_reader_utf8_lossy<R: Read>(mut reader: R) -> crate::de::Result<Document> {
+        Self::decode(&mut reader, true)
     }
 }
 
-pub struct Entry<'a> {
-    inner: linked_hash_map::Entry<'a, String, Bson>,
+/// A view into a single entry in a map, which may either be vacant or occupied.
+///
+/// This enum is constructed from the entry method on HashMap.
+pub enum Entry<'a> {
+    /// An occupied entry.
+    Occupied(OccupiedEntry<'a>),
+
+    /// A vacant entry.
+    Vacant(VacantEntry<'a>),
 }
 
 impl<'a> Entry<'a> {
+    /// Returns a reference to this entry's key.
     pub fn key(&self) -> &str {
+        match self {
+            Self::Vacant(v) => v.key(),
+            Self::Occupied(o) => o.key(),
+        }
+    }
+
+    fn into_indexmap_entry(self) -> indexmap::map::Entry<'a, String, Bson> {
+        match self {
+            Self::Occupied(o) => indexmap::map::Entry::Occupied(o.inner),
+            Self::Vacant(v) => indexmap::map::Entry::Vacant(v.inner),
+        }
+    }
+
+    /// Inserts the given default value in the entry if it is vacant and returns a mutable reference
+    /// to it. Otherwise a mutable reference to an already existent value is returned.
+    pub fn or_insert(self, default: Bson) -> &'a mut Bson {
+        self.into_indexmap_entry().or_insert(default)
+    }
+
+    /// Inserts the result of the `default` function in the entry if it is vacant and returns a
+    /// mutable reference to it. Otherwise a mutable reference to an already existent value is
+    /// returned.
+    pub fn or_insert_with<F: FnOnce() -> Bson>(self, default: F) -> &'a mut Bson {
+        self.into_indexmap_entry().or_insert_with(default)
+    }
+}
+
+/// A view into a vacant entry in a [Document]. It is part of the [Entry] enum.
+pub struct VacantEntry<'a> {
+    inner: indexmap::map::VacantEntry<'a, String, Bson>,
+}
+
+impl<'a> VacantEntry<'a> {
+    /// Gets a reference to the key that would be used when inserting a value through the
+    /// [VacantEntry].
+    fn key(&self) -> &str {
         self.inner.key()
     }
-
-    pub fn or_insert(self, default: Bson) -> &'a mut Bson {
-        self.inner.or_insert(default)
-    }
-
-    pub fn or_insert_with<F: FnOnce() -> Bson>(self, default: F) -> &'a mut Bson {
-        self.inner.or_insert_with(default)
-    }
 }
 
-impl From<LinkedHashMap<String, Bson>> for Document {
-    fn from(tree: LinkedHashMap<String, Bson>) -> Document {
-        Document { inner: tree }
-    }
+/// A view into an occupied entry in a [Document]. It is part of the [Entry] enum.
+pub struct OccupiedEntry<'a> {
+    inner: indexmap::map::OccupiedEntry<'a, String, Bson>,
 }
 
-pub struct DocumentVisitor {
-    marker: PhantomData<Document>,
-}
-
-impl DocumentVisitor {
-    #[allow(clippy::new_without_default)]
-    pub fn new() -> DocumentVisitor {
-        DocumentVisitor {
-            marker: PhantomData,
-        }
-    }
-}
-
-impl<'de> Visitor<'de> for DocumentVisitor {
-    type Value = Document;
-
-    fn expecting(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        write!(f, "expecting ordered document")
-    }
-
-    #[inline]
-    fn visit_unit<E>(self) -> Result<Document, E>
-    where
-        E: de::Error,
-    {
-        Ok(Document::new())
-    }
-
-    #[inline]
-    fn visit_map<V>(self, mut visitor: V) -> Result<Document, V::Error>
-    where
-        V: MapAccess<'de>,
-    {
-        let mut inner = match visitor.size_hint() {
-            Some(size) => LinkedHashMap::with_capacity(size),
-            None => LinkedHashMap::new(),
-        };
-
-        while let Some((key, value)) = visitor.next_entry()? {
-            inner.insert(key, value);
-        }
-
-        Ok(inner.into())
+impl<'a> OccupiedEntry<'a> {
+    /// Gets a reference to the key in the entry.
+    pub fn key(&self) -> &str {
+        self.inner.key()
     }
 }
 
