@@ -1,66 +1,46 @@
-use std::{error, fmt, fmt::Display, io};
+use std::{error, fmt, fmt::Display, io, sync::Arc};
 
 use serde::ser;
 
 use crate::bson::Bson;
 
 /// Possible errors that can arise during encoding.
-#[derive(Debug)]
+#[derive(Clone, Debug)]
 #[non_exhaustive]
 pub enum Error {
     /// A [`std::io::Error`](https://doc.rust-lang.org/std/io/struct.Error.html) encountered while serializing.
-    IoError(io::Error),
+    Io(Arc<io::Error>),
 
     /// A key could not be serialized to a BSON string.
-    InvalidMapKeyType {
-        /// The value that could not be used as a key.
-        key: Bson,
-    },
-
-    /// An invalid string was specified.
-    InvalidCString(String),
+    InvalidDocumentKey(Bson),
 
     /// A general error that occurred during serialization.
-    /// See: https://docs.rs/serde/1.0.110/serde/ser/trait.Error.html#tymethod.custom
+    /// See: <https://docs.rs/serde/1.0.110/serde/ser/trait.Error.html#tymethod.custom>
+    #[non_exhaustive]
     SerializationError {
         /// A message describing the error.
         message: String,
     },
 
-    #[cfg(not(feature = "u2i"))]
-    /// Returned when serialization of an unsigned integer was attempted. BSON only supports
-    /// 32-bit and 64-bit signed integers.
-    ///
-    /// To allow serialization of unsigned integers as signed integers, enable the "u2i" feature
-    /// flag.
-    UnsupportedUnsignedType,
-
-    #[cfg(feature = "u2i")]
     /// An unsigned integer type could not fit into a signed integer type.
-    UnsignedTypesValueExceedsRange(u64),
+    UnsignedIntegerExceededRange(u64),
 }
 
 impl From<io::Error> for Error {
     fn from(err: io::Error) -> Error {
-        Error::IoError(err)
+        Error::Io(Arc::new(err))
     }
 }
 
 impl fmt::Display for Error {
     fn fmt(&self, fmt: &mut fmt::Formatter) -> fmt::Result {
         match *self {
-            Error::IoError(ref inner) => inner.fmt(fmt),
-            Error::InvalidMapKeyType { ref key } => write!(fmt, "Invalid map key type: {}", key),
-            Error::InvalidCString(ref string) => {
-                write!(fmt, "cstrings cannot contain null bytes: {:?}", string)
-            }
+            Error::Io(ref inner) => inner.fmt(fmt),
+            Error::InvalidDocumentKey(ref key) => write!(fmt, "Invalid map key type: {}", key),
             Error::SerializationError { ref message } => message.fmt(fmt),
-            #[cfg(not(feature = "u2i"))]
-            Error::UnsupportedUnsignedType => fmt.write_str("BSON does not support unsigned type"),
-            #[cfg(feature = "u2i")]
-            Error::UnsignedTypesValueExceedsRange(value) => write!(
+            Error::UnsignedIntegerExceededRange(value) => write!(
                 fmt,
-                "BSON does not support unsigned types.
+                "BSON does not support unsigned integers.
                  An attempt to serialize the value: {} in a signed type failed due to the value's \
                  size.",
                 value
@@ -72,7 +52,7 @@ impl fmt::Display for Error {
 impl error::Error for Error {
     fn cause(&self) -> Option<&dyn error::Error> {
         match *self {
-            Error::IoError(ref inner) => Some(inner),
+            Error::Io(ref inner) => Some(inner.as_ref()),
             _ => None,
         }
     }
